# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration record table load module."""

from datetime import datetime

from ....load.ids import generate_recid, generate_uuid, pid_pk
from ....load.postgresql.bulk.generators import TableGenerator
from ....state import STATE
from ...models.pids import PersistentIdentifier
from ...models.records import (
    RDMParentMetadata,
    RDMRecordFile,
    RDMRecordMediaFile,
    RDMRecordMetadata,
)
from .parents import generate_parent_rows
from .references import CommunitiesReferencesMixin
from .utils import InsertRecordFiles


def generate_record_uuid(data):
    """Generate record uuid if not present."""
    _id = data["record"].get("id")
    return _id if _id else generate_uuid(None)


class RDMRecordTableGenerator(TableGenerator, CommunitiesReferencesMixin):
    """RDM Record and related tables load."""

    def __init__(self):
        """Constructor."""
        super().__init__(
            tables=[
                PersistentIdentifier,
                RDMParentMetadata,
                RDMRecordMetadata,
            ],
            post_load_hooks=[
                InsertRecordFiles(RDMRecordMetadata, RDMRecordFile),
                InsertRecordFiles(
                    RDMRecordMetadata,
                    RDMRecordMediaFile,
                    bucket_fk="media_bucket_id",
                ),
            ],
            pks=[
                ("record.id", generate_record_uuid),
                ("parent.id", generate_uuid),
                ("record.json.pid", generate_recid),
                ("parent.json.pid", generate_recid),
                ("record.parent_id", lambda d: d["parent"]["id"]),
            ],
        )

    def _generate_rows(self, data, **kwargs):
        """Generates rows for a record."""
        now = datetime.utcnow().isoformat()
        parent = data["parent"]
        record = data.get("record")

        if not record:
            return

        record_pid = record["json"]["pid"]

        # Handle parent
        state_parent = STATE.PARENTS.get(parent["json"]["id"])
        if not state_parent:
            STATE.PARENTS.add(
                parent["json"]["id"],  # recid
                {
                    "id": parent["id"],
                    "latest_index": record["index"],
                    "latest_id": record["id"],
                    "communities": parent.get("communities", {}).get("ids", []),
                },
            )
            yield from generate_parent_rows(parent)
        else:
            if state_parent.get("latest_index") < record["index"]:
                STATE.PARENTS.update(
                    parent["json"]["id"],
                    {
                        "latest_index": record["index"],
                        "latest_id": record["id"],
                    },
                )
        parent_id = state_parent["id"] if state_parent else record["parent_id"]

        # record
        STATE.RECORDS.add(
            record["json"]["id"],  # recid
            {
                "index": record["index"],
                "id": record["id"],  # uuid
                "parent_id": parent_id,  # parent uuid
                "fork_version_id": record["version_id"],
                "pids": record["json"]["pids"],
            },
        )

        yield RDMRecordMetadata(
            id=record["id"],
            json=record["json"],
            created=record["created"],
            updated=record["updated"],
            version_id=record["version_id"],
            index=record["index"],
            bucket_id=record["bucket_id"],
            media_bucket_id=record.get("media_bucket_id"),
            parent_id=parent_id,
            deletion_status="P",
        )
        # recid
        yield PersistentIdentifier(
            id=record_pid["pk"],
            pid_type=record_pid["pid_type"],
            pid_value=record["json"]["id"],
            status=record_pid["status"],
            object_type=record_pid["obj_type"],
            object_uuid=record["id"],
            created=now,
            updated=now,
        )
        # DOI
        if "doi" in record["json"]["pids"]:
            yield PersistentIdentifier(
                id=pid_pk(),
                pid_type="doi",
                pid_value=record["json"]["pids"]["doi"]["identifier"],
                status="R",
                object_type="rec",
                object_uuid=record["id"],
                created=now,
                updated=now,
            )
        # OAI
        if "oai" in record["json"]["pids"]:
            yield PersistentIdentifier(
                id=pid_pk(),
                pid_type="oai",
                pid_value=record["json"]["pids"]["oai"]["identifier"],
                status="R",
                object_type="rec",
                object_uuid=record["id"],
                created=now,
                updated=now,
            )

    def _resolve_references(self, data, **kwargs):
        """Resolve references e.g communities slug names."""
        if "record" in data:
            # resolve parent communities slug
            parent = data["parent"]
            communities = parent["json"].get("communities")
            if communities:
                self.resolve_communities(communities)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration PostgreSQL bulk generators module."""

from .existing import ExistingDataTableGenerator
from .single import SingleTableGenerator
from .table import TableGenerator

__all__ = (
    "ExistingDataTableGenerator",
    "SingleTableGenerator",
    "TableGenerator",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""OAuth2Server applications actions module."""

from dataclasses import dataclass

from .....actions import LoadAction, LoadData
from .....load.postgresql.transactions.operations import Operation, OperationType
from ....models.oauth import ServerClient


@dataclass
class OAuthApplicationData(LoadData):
    """OAuth application action data."""

    client: dict


class OAuthApplicationCreateAction(LoadAction):
    """Create a application auth."""

    name = "oauth-application-create"
    data_cls = OAuthApplicationData

    def _generate_rows(self, **kwargs):
        """Generates rows for a new oauth token."""
        yield Operation(OperationType.INSERT, ServerClient, self.data.client)


class OAuthApplicationUpdateAction(LoadAction):
    """Update a application auth.

    Note a client secret reset can be treated as an update.
    """

    name = "oauth-application-update"
    data_cls = OAuthApplicationData

    def _generate_rows(self, **kwargs):
        """Generates rows for a new oauth token."""
        yield Operation(OperationType.UPDATE, ServerClient, self.data.client)


class OAuthApplicationDeleteAction(LoadAction):
    """Delete a application auth."""

    name = "oauth-application-delete"
    data_cls = OAuthApplicationData

    def _generate_rows(self, **kwargs):
        """Generates rows for a new oauth token."""
        yield Operation(OperationType.DELETE, ServerClient, self.data.client)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""JSON transformation tests."""

from invenio_rdm_migrator.transform import JSONTransformMixin


def test_load_json_fields():
    data = {"json_field": '{"some": "value", "another":[1,2,3]}', "not_json": "test"}

    class TestTransform(JSONTransformMixin):
        """Test JSON transform."""

        pass

    transform = TestTransform()
    transform._load_json_fields(data=data, fields=["json_field"])

    assert data == {
        "json_field": {"some": "value", "another": [1, 2, 3]},
        "not_json": "test",
    }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration PostgreSQL models module."""

from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import DeclarativeBase, MappedAsDataclass


class Model(MappedAsDataclass, DeclarativeBase):
    """subclasses will be converted to dataclasses."""

    type_annotation_map = {
        dict: JSONB,
    }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Identity transaction transformation tests."""

from dataclasses import dataclass

import pytest

from invenio_rdm_migrator.actions import LoadAction, LoadData, TransformAction
from invenio_rdm_migrator.extract import Tx
from invenio_rdm_migrator.transform import BaseTxTransform
from invenio_rdm_migrator.transform.errors import MultipleActionMatches, NoActionMatch


@pytest.fixture()
def tx(scope="function"):
    return Tx(
        id=1,
        operations=[
            {
                "table": "records",
                "op": "I",
                "after": {"id": "abc", "json": {"title": "Test title"}},
            },
            {"table": "files", "op": "I", "after": {"key": "data.zip"}},
            {"table": "files", "op": "I", "after": {"key": "article.pdf"}},
            {"table": "files", "op": "I", "after": {"key": "figure.png"}},
        ],
    )


@dataclass
class TestLoadData(LoadData):
    record: dict
    files: list[dict]


class TestLoadAction(LoadAction):
    """Test load action."""

    name = "test-load"

    data_cls = TestLoadData

    def _generate_rows(self, **kwargs):  # pragma: no cover
        """Yield generated rows."""
        yield


class TestTransformAction(TransformAction):
    """Test transform action."""

    name = "test-transform"
    load_cls = TestLoadAction

    @classmethod
    def matches_action(cls, tx: Tx):  # pragma: no cover
        """Checks if the data matches with that required by the action."""
        record_ops = [
            o for o in tx.operations if o["table"] == "records" and o["op"] == "I"
        ]
        has_one_record_insert = len(record_ops) == 1
        has_files_insert = any(
            o for o in tx.operations if o["table"] == "files" and o["op"] == "I"
        )
        return has_one_record_insert and has_files_insert

    def _transform_data(self):  # pragma: no cover
        """Transforms the data and returns an instance of the ``data_cls``."""
        record = next(o["after"] for o in self.tx.operations if o["table"] == "records")
        record["json"]["titles"] = [record["json"].pop("title")]
        return dict(
            tx=self.tx,
            record=record,
            files=[o["after"] for o in self.tx.operations if o["table"] == "files"],
        )


class TestTxTransform(BaseTxTransform):
    actions = [
        TestTransformAction,
    ]


def test_transform_returns_load_action(tx):
    transform = TestTxTransform()
    load_action = transform._transform(tx)
    assert isinstance(load_action, TestLoadAction)
    assert isinstance(load_action.data, TestLoadData)
    assert load_action.data.tx == tx
    assert load_action.data.record == {"id": "abc", "json": {"titles": ["Test title"]}}
    assert load_action.data.files == [
        # NOTE: Order doesn't really matter here
        {"key": "data.zip"},
        {"key": "article.pdf"},
        {"key": "figure.png"},
    ]


def test_multiple_action_matches(tx):
    class MultipleTxTransform(BaseTxTransform):
        actions = [
            TestTransformAction,
            TestTransformAction,
        ]

    transform = MultipleTxTransform()
    with pytest.raises(MultipleActionMatches):
        transform._transform(tx)


def test_no_action_matches():
    tx = Tx(
        id=1,
        operations=[
            {"table": "random", "op": "I", "after": {"some": "value"}},
        ],
    )
    transform = TestTxTransform()
    with pytest.raises(NoActionMatch):
        transform._transform(tx)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test fixtures and utilities."""

import pytest
import sqlalchemy as sa
from sqlalchemy.orm import Session


@pytest.fixture(scope="session")
def db_uri():
    """Database URI."""
    return "postgresql+psycopg://invenio:invenio@localhost:5432/invenio"


@pytest.fixture(scope="session")
def engine(db_uri):
    """SQLAlchemy engine."""
    return sa.create_engine(db_uri)


@pytest.fixture(scope="function")
def session(engine):
    """SQLAlchemy session."""
    conn = engine.connect()
    transaction = conn.begin()

    session = Session(bind=conn, join_transaction_mode="create_savepoint")

    yield session

    session.close()
    transaction.rollback()
    conn.close()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration transform module."""

from .base import Entry, Transform, drop_nones
from .dates import DatetimeMixin
from .encrypt import EncryptMixin
from .identity import IdentityTransform
from .json import JSONTransformMixin
from .transactions import BaseTxTransform

__all__ = (
    "drop_nones",
    "BaseTxTransform",
    "DatetimeMixin",
    "EncryptMixin",
    "Entry",
    "IdentityTransform",
    "JSONTransformMixin",
    "Transform",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Dataclasses OAuth models to generate table rows."""
from dataclasses import InitVar
from datetime import datetime

from sqlalchemy.orm import Mapped, mapped_column

from ...load.postgresql.models import Model


class RemoteAccount(Model):
    """OAuth client remote account dataclass model."""

    __tablename__: InitVar[str] = "oauthclient_remoteaccount"

    id: Mapped[int] = mapped_column(primary_key=True)
    user_id: Mapped[int]
    client_id: Mapped[str]
    extra_data: Mapped[dict]
    created: Mapped[datetime]
    updated: Mapped[datetime]


class RemoteToken(Model):
    """OAuth client remote token dataclass model."""

    __tablename__: InitVar[str] = "oauthclient_remotetoken"

    id_remote_account: Mapped[int] = mapped_column(primary_key=True)
    token_type: Mapped[str] = mapped_column(primary_key=True)
    access_token: Mapped[str]
    secret: Mapped[str]
    created: Mapped[datetime]
    updated: Mapped[datetime]


class ServerClient(Model):
    """OAuth server client dataclass model."""

    __tablename__: InitVar[str] = "oauth2server_client"

    name: Mapped[str]
    client_id: Mapped[str] = mapped_column(primary_key=True)
    description: Mapped[str] = mapped_column(nullable=True)
    website: Mapped[str] = mapped_column(nullable=True)
    user_id: Mapped[int] = mapped_column(nullable=True)
    client_secret: Mapped[str]  # it is text, migrated as is
    _redirect_uris: Mapped[str] = mapped_column(nullable=True)
    _default_scopes: Mapped[str] = mapped_column(nullable=True)
    is_internal: Mapped[bool] = mapped_column(nullable=True, default=False)
    is_confidential: Mapped[bool] = mapped_column(nullable=True, default=True)


class ServerToken(Model):
    """OAuth server token dataclass model."""

    __tablename__: InitVar[str] = "oauth2server_token"

    id: Mapped[int] = mapped_column(primary_key=True)
    client_id: Mapped[str]
    user_id: Mapped[int] = mapped_column(nullable=True)
    access_token: Mapped[str] = mapped_column(nullable=True)
    refresh_token: Mapped[str] = mapped_column(nullable=True)
    expires: Mapped[datetime] = mapped_column(nullable=True)
    _scopes: Mapped[str] = mapped_column(nullable=True)
    token_type: Mapped[str] = mapped_column(nullable=True, default="bearer")
    is_personal: Mapped[bool] = mapped_column(nullable=True, default=False)
    is_internal: Mapped[bool] = mapped_column(nullable=True, default=False)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Identity transformation tests."""

import pytest
from sqlalchemy_utils.types.encrypted.encrypted_type import AesEngine

from invenio_rdm_migrator.transform import EncryptMixin, Transform


@pytest.fixture(scope="function")
def transform_with_mixin():
    """Test instance of a transform class with identity mixin"""

    class FixtureTransform(EncryptMixin, Transform):
        """Transform fixture class."""

        def _token(self, entry):
            return self.re_encrypt(entry["token"])

        def _transform(self, entry):
            """Transform entry."""
            return {
                "token": self._token(entry),
            }

    return FixtureTransform()


def test_encrypt_transform(secret_keys_state, transform_with_mixin):
    token = "itsasecret"

    encrypt_engine = AesEngine()
    encrypt_engine._update_key(bytes("OLDKEY", "utf-8"))
    encrypt_engine._set_padding_mechanism()

    encrypted_token = encrypt_engine.encrypt(token)

    t_item = transform_with_mixin._transform({"token": encrypted_token})
    t_token = t_item["token"]

    assert t_token != token != encrypted_token

    decrypt_engine = AesEngine()
    decrypt_engine._update_key(bytes("NEWKEY", "utf-8"))
    decrypt_engine._set_padding_mechanism()

    assert decrypt_engine.decrypt(t_token) == token



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Null extract tests."""

from types import GeneratorType

import pytest

from invenio_rdm_migrator.extract.null import NullExtract


def test_run_method_does_not_raise_errors():
    extract = NullExtract()
    try:
        next(extract.run())
    except Exception as e:
        pytest.fail(f"Unexpected exception: {e}")


def test_next_method_returns_none():
    extract = NullExtract()
    assert next(extract.run()) is None


def test_run_returns_iterator():
    extract = NullExtract()
    assert isinstance(extract.run(), GeneratorType)


def test_generator_object_only_returns_once():
    extract = NullExtract()
    gen = extract.run()
    next(gen)
    with pytest.raises(StopIteration):
        next(gen)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Base action module.

In the table generator approach, the E T L classes contain the logic to process the
entries. When using actions the responsibility is shifted to these classes, making
the E T L instances a lightweight container that "calls" the corresponding method on
an action. Being the latter the one implementing the logic.
"""

from abc import ABC, abstractclassmethod, abstractmethod
from dataclasses import dataclass
from typing import ClassVar, Optional

import sqlalchemy.orm as orm

from invenio_rdm_migrator.extract.transactions import Tx

from ..logging import Logger
from ..transform import DatetimeMixin
from ..utils import dict_set


class Action:
    """Base action."""

    name = None

    def __init__(self):
        """Constructor."""
        assert self.name  # it must have a name


@dataclass
class LoadData:
    """Load action data."""

    tx: Optional[Tx]


class LoadAction(Action, ABC):
    """Load action.

    This generates the corresponding SQL operations to perform.
    """

    data_cls: ClassVar[type[LoadData]] = None
    pks = []

    def __init__(self, data: dict):
        """Constructor.

        :param pks: a triplet with the attribute, the key and the function.
        """
        assert data is not None
        data.setdefault("tx", None)
        self.data = self.data_cls(**data)
        super().__init__()

    def _generate_pks(self, session: orm.Session = None):
        for attr_name, path, pk_func in self.pks:
            try:
                attr = getattr(self.data, attr_name)
                if attr:
                    dict_set(attr, path, pk_func(attr))
            except AttributeError:
                logger = Logger.get_logger()
                logger.exception(f"Attribute {attr_name} not found", exc_info=1)
            except KeyError:
                logger = Logger.get_logger()
                logger.exception(f"Path {path} not found on record", exc_info=1)

    def _resolve_references(self, session: orm.Session = None, **kwargs):
        """Resolve references e.g communities slug names."""
        pass

    @abstractmethod
    def _generate_rows(self, session: orm.Session = None, **kwargs):
        """Yield generated rows."""

    def prepare(self, session: orm.Session, **kwargs):
        """Generate the SQL statements required to persist the action.

        Any required data should be part of the instance attributes.
        """
        # is_db_empty would come in play and make _generate_pks optional
        self._generate_pks(session=session)
        # resolve entry references
        self._resolve_references(session=session)
        yield from self._generate_rows(session=session, **kwargs)


class TransformAction(Action, DatetimeMixin, ABC):
    """Transform action.

    Detects/matches an action type based on transaction data and transforms the data
    into a target ``LoadAction``.
    """

    load_cls: ClassVar[type[LoadAction]] = None

    def __init__(self, tx: Tx):
        """Constructor."""
        assert self.load_cls is not None
        self.tx: Tx = tx
        super().__init__()

    def transform(self):
        """Transforms an action."""
        transformed_data = self._transform_data()
        transformed_data.pop("tx_id", None)
        transformed_data["tx"] = self.tx
        return self.load_cls(transformed_data)

    @abstractclassmethod
    def matches_action(cls, tx: Tx):  # pragma: no cover
        """Checks if the data matches with that required by the action."""

    @abstractmethod
    def _transform_data(self):  # pragma: no cover
        """Transforms the data and returns a dictionary."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""IDs generation tests."""

from uuid import UUID

from invenio_rdm_migrator.load.ids import (
    generate_pk,
    generate_recid,
    generate_uuid,
    pid_pk,
)


def test_valid_input():
    uuid = generate_uuid(None)
    assert isinstance(uuid, str)
    UUID(uuid)  # will raise ValueError if its not a UUID


def test_max_pid_pk(state):
    """Test max pid pk value changes."""
    # check state is empty
    assert not state.VALUES.get("max_pid_pk")
    # create pid
    val = pid_pk()
    # check state has been created
    assert val == state.VALUES.get("max_pid_pk")["value"]
    # create another pid
    val_inc = pid_pk()
    # check the state has been incremented
    inc_val = val + 1
    assert inc_val == val_inc
    assert val_inc == state.VALUES.get("max_pid_pk")["value"]


def test_generate_pk_returns_integer():
    assert isinstance(generate_pk(None), int)


def test_recid_default_status():
    rec_id = generate_recid({})
    assert rec_id["status"] == "R"


def test_recid_custom_status():
    rec_id = generate_recid({}, "C")
    assert rec_id["status"] == "C"


def test_recid_return_value_types():
    rec_id = generate_recid({})
    assert isinstance(rec_id["pk"], int)
    assert rec_id["obj_type"] == "rec"
    assert rec_id["pid_type"] == "recid"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM records table generators references."""

from invenio_rdm_migrator.logging import Logger

from ....state import STATE


class CommunitiesReferencesMixin:
    """Communities reference resolver."""

    def resolve_communities(self, communities):
        """Translates communities slugs to ids.

        Assumes the the table generator has a communities_state attribute.
        """
        logger = Logger.get_logger()
        default_slug = communities.get("default")
        if default_slug:
            default_comm = STATE.COMMUNITIES.get(default_slug)
            if "id" in default_comm:
                communities["default"] = default_comm["id"]
            else:
                # raise error without correct default community?
                logger.warning(
                    f"Community with slug [{default_slug}] was not in state [{communities}]"
                )
                communities.pop("default", None)
                communities["ids"].remove(default_slug)

        communities_slugs = communities.get("ids", [])
        _ids = []
        for slug in communities_slugs:
            comm = STATE.COMMUNITIES.get(slug)
            if "id" in comm:
                _ids.append(comm["id"])
            else:
                logger.warning(
                    f"Community with slug [{slug}] was not in state [{communities}]"
                )

        communities["ids"] = _ids


class PIDsReferencesMixin:
    """PIDs reference resolver."""

    def resolve_draft_pids(self, draft):
        """Enforce record pids to draft."""
        if not draft:
            return

        # some legacy records have different pid value in deposit than record
        # however _deposit.pid.value would contain the correct one
        # if it is not legacy we get it from the current field (json.id)
        recid = draft["json"]["id"]
        forked_published = STATE.RECORDS.get(recid)
        if forked_published:
            pids = draft["json"]["pids"]
            has_draft_external_doi = pids.get("doi", {}).get("provider") == "external"
            if has_draft_external_doi:
                # then keep the draft external value as it might be there for
                # updating the existing value. Update the draft only with `oai`
                pids["oai"] = forked_published["pids"]["oai"]
            else:
                # enfore published record pids to draft
                pids = forked_published["pids"]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Pytest requests configuration."""

from copy import deepcopy

import pytest


@pytest.fixture(scope="function")
def transformed_record_entry():
    """Record entry after being transformed."""
    return {
        "parent": {
            "created": "2023-01-01 12:00:00.00000",
            "updated": "2023-01-31 12:00:00.00000",
            "version_id": 1,
            "json": {
                "id": "12345677",
                "communities": {"ids": ["comm", "other-comm"], "default": "comm"},
            },
        },
        "record": {
            "id": "2d6970ea-602d-4e8b-a918-063a59823386",
            "created": "2023-01-01 12:00:00.00000",
            "updated": "2023-01-31 12:00:00.00000",
            "version_id": 1,
            "index": 1,
            "bucket_id": "bur3c0rd-1234-abcd-1ab2-1234abcd56ef",
            "json": {
                "id": "12345678",
                "pids": {
                    "oai": {
                        "provider": "oai",
                        "identifier": "oai:zenodo.org:12345678",
                    },
                    "doi": {
                        "client": "datacite",
                        "provider": "datacite",
                        "identifier": "10.5281/zenodo.12345678",
                    },
                },
            },
        },
    }


@pytest.fixture(scope="function")
def transformed_record_entry_pks(transformed_record_entry):
    """Record data as after being transformed.

    It also includes generated PKs and resolved references.
    """
    data = deepcopy(transformed_record_entry)
    data["parent"]["id"] = "12345678-abcd-1a2b-3c4d-123abc456def"
    data["record"]["parent_id"] = "12345678-abcd-1a2b-3c4d-123abc456def"
    data["parent"]["json"]["pid"] = {
        "pk": 1_000_000,  # pk, not pid value
        "obj_type": "rec",
        "pid_type": "recid",
        "status": "R",
    }
    data["parent"]["json"]["communities"] = {
        "ids": [
            "12345678-abcd-1a2b-3c4d-123abc456def",
            "12345678-abcd-1a2b-3c4d-123abc123abc",
        ],
        "default": "12345678-abcd-1a2b-3c4d-123abc456def",
    }
    data["record"]["json"]["pid"] = {
        "pk": 1_000_001,  # pk, not pid value
        "obj_type": "rec",
        "pid_type": "recid",
        "status": "R",
    }

    return data


@pytest.fixture(scope="function")
def transformed_draft_entry():
    """Draft entry after being transformed."""
    return {
        "parent": {
            "created": "2023-01-01 12:00:00.00000",
            "updated": "2023-01-31 12:00:00.00000",
            "version_id": 1,
            "json": {
                "id": "12345677",
                "communities": {"ids": ["comm", "other-comm"], "default": "comm"},
            },
        },
        "draft": {
            "id": "2d6970ea-602d-4e8b-a918-063a59823386",
            "created": "2023-01-01 12:00:00.00000",
            "updated": "2023-01-31 12:00:00.00000",
            "version_id": 1,
            "index": 1,
            "bucket_id": "b0c73700-1234-abcd-1ab2-1234abcd56ef",
            "json": {
                "id": "12345678",
                "pids": {
                    "doi": {
                        "provider": "external",
                        "identifier": "10.1234/foo",
                    },
                },
            },
            "expires_at": "2024-01-01 12:00:00.00000",
            "fork_version_id": None,
        },
    }


@pytest.fixture(scope="function")
def transformed_draft_entry_pks(transformed_draft_entry):
    """Draft data as after being transformed.

    It also includes generated PKs and resolved references.
    """
    data = deepcopy(transformed_draft_entry)
    data["parent"]["id"] = "12345678-abcd-1a2b-3c4d-123abc456def"
    data["draft"]["parent_id"] = "12345678-abcd-1a2b-3c4d-123abc456def"
    data["parent"]["json"]["pid"] = {
        "pk": 1_000_000,
        "obj_type": "rec",
        "pid_type": "recid",
        "status": "N",
    }
    data["parent"]["json"]["communities"] = {
        "ids": [
            "12345678-abcd-1a2b-3c4d-123abc456def",
            "12345678-abcd-1a2b-3c4d-123abc123abc",
        ],
        "default": "12345678-abcd-1a2b-3c4d-123abc456def",
    }
    data["draft"]["json"]["pid"] = {
        "pk": 1_000_001,
        "obj_type": "rec",
        "pid_type": "recid",
        "status": "N",
    }
    return data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration requests transform interfaces."""

from abc import abstractmethod

from ...transform import Entry, Transform


class RequestTransform(Transform):
    """Base class for request data transformation."""

    @abstractmethod
    def _request(self, entry):
        """Transform the request."""
        pass

    def _transform(self, entry):
        """Transform a single entry."""
        return self._request(entry)


class RequestEntry(Entry):
    """Transform a single request entry."""

    @abstractmethod
    def _created(self, entry):
        """Returns the creation date of the request."""
        pass

    @abstractmethod
    def _updated(self, entry):
        """Returns the update date of the request."""
        pass

    @abstractmethod
    def _version_id(self, entry):
        """Returns the version id of the request.

        {Elastic,Open}search will fail with values lower than 1.
        """
        pass

    @abstractmethod
    def _json(self, entry):
        """Returns the request json body."""
        pass

    @abstractmethod
    def _number(self, entry):
        """Returns the request number."""
        pass

    @abstractmethod
    def _expires_at(self, entry):
        """Returns the request expiration date."""
        pass

    def transform(self, entry):
        """Transform a request single entry."""
        return {
            "created": self._created(entry),
            "updated": self._updated(entry),
            "version_id": self._version_id(entry),
            "json": self._json(entry),
            "number": self._number(entry),
            "expires_at": self._expires_at(entry),
        }


class InclusionRequestEntry(RequestEntry):
    """Inclusion request entry."""

    @abstractmethod
    def _title(self, entry):
        """Returns the request title."""
        pass

    @abstractmethod
    def _topic(self, entry):
        """Returns the request topic."""
        pass

    @abstractmethod
    def _receiver(self, entry):
        """Returns the request receiver."""
        pass

    @abstractmethod
    def _created_by(self, entry):
        """Returns the request creation reference."""
        pass

    def _json(self, entry):
        """Returns the request json body."""
        return {
            "type": "community-inclusion",
            "title": self._title(entry),
            "topic": self._topic(entry),
            "status": "submitted",
            "receiver": self._receiver(entry),
            "created_by": self._created_by(entry),
            "$schema": "local://requests/request-v1.0.0.json",
        }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Logging tests."""


from invenio_rdm_migrator.logging import Logger

# NOTE this tests will actually print so they will display in the console when running
# pytest. However, they will pass.


def test_log_info_to_stdout(capsys, tmp_path):
    Logger.initialize(tmp_path)
    logger = Logger.get_logger()

    logger.info("safe to ignore: Test message")
    captured = capsys.readouterr()
    assert "Test message" in captured.err


def test_log_error_to_file(tmp_path):
    Logger.initialize(tmp_path)
    logger = Logger.get_logger()

    with open(tmp_path / "error.log", "r") as f:
        logger.error("safe to ignore: Test error message")
        assert "Test error message" in f.read()

        logger.info("safe to ignore: Test info message")
        assert not "Test info message" in f.read()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM records table generators."""

from .deleted_records import RDMDeletedRecordTableGenerator
from .drafts import RDMDraftTableGenerator
from .records import RDMRecordTableGenerator
from .versions import RDMVersionStateTableGenerator

__all__ = (
    "RDMDraftTableGenerator",
    "RDMRecordTableGenerator",
    "RDMVersionStateTableGenerator",
    "RDMDeletedRecordTableGenerator",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""PostgreSQL Execute load."""

import sqlalchemy as sa
from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from ....logging import FailedTxLogger, Logger
from ...base import Load
from .operations import OperationType


class PostgreSQLTx(Load):
    """PostgreSQL COPY load."""

    def __init__(
        self, db_uri, _session=None, dry=True, raise_on_db_error=False, **kwargs
    ):
        """Constructor."""
        self.db_uri = db_uri
        self.dry = dry
        self.raise_on_db_error = raise_on_db_error
        self._session = _session

    @property
    def session(self):
        """DB session."""
        if self._session is None:
            session_kwargs = dict(bind=create_engine(self.db_uri))
            if self.dry:
                session_kwargs["join_transaction_mode"] = "create_savepoint"
            self._session = Session(**session_kwargs)
        return self._session

    def _cleanup(self, db=False):
        """No cleanup."""
        logger = Logger.get_logger()
        logger.debug("PostgreSQLExecute does not implement _cleanup()")
        pass

    def _load(self, transactions):
        """Performs the operations of a group transaction."""
        logger = Logger.get_logger()
        failed_tx_logger = FailedTxLogger.get_logger()
        exec_kwargs = dict(execution_options={"synchronize_session": False})

        outer_trans = None
        if self.dry:
            outer_trans = self.session.begin()
        try:
            for action in transactions:
                with self.session.no_autoflush:
                    nested_trans = self.session.begin_nested()
                    try:
                        for op in action.prepare(session=self.session):
                            if op.type == OperationType.INSERT:
                                row = op.as_row_dict()
                                logger.info(f"INSERT {op.model}: {row}")
                                self.session.execute(
                                    sa.insert(op.model),
                                    [row],
                                    **exec_kwargs,
                                )
                            elif op.type == OperationType.DELETE:
                                logger.info(f"DELETE {op.model}: {op.data}")
                                self.session.execute(
                                    sa.delete(op.model).where(*op.pk_clauses),
                                    **exec_kwargs,
                                )
                            elif op.type == OperationType.UPDATE:
                                row = op.as_row_dict()
                                logger.info(f"UDPATE {op.model}: {op.data}")
                                self.session.execute(
                                    sa.update(op.model),
                                    [row],
                                    **exec_kwargs,
                                )
                            self.session.flush()
                        nested_trans.commit()
                    except Exception:
                        logger.exception(
                            f"Could not load {action.data} ({action.name})",
                            exc_info=True,
                        )
                        failed_tx_logger.exception(
                            "Failed processing transaction",
                            extra={"tx": action.data},
                            exc_info=True,
                        )
                        nested_trans.rollback()
                        if self.raise_on_db_error:
                            raise
        except Exception:
            logger.exception("Transactions load failed", exc_info=True)
            failed_tx_logger.exception("Failed transaction", exc_info=True)
            if self.raise_on_db_error:
                # NOTE: the "finally" block below will run before this "raise"
                raise
        finally:
            if self.dry and outer_trans:
                outer_trans.rollback()

    def run(self, entries, cleanup=False):
        """Load entries."""
        self._load(entries)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration transform interfaces."""

from .base import Transform


class IdentityTransform(Transform):
    """Transform class to yield the received item without change."""

    def _transform(self, entry):
        """Transform entry."""
        return entry



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""GitHub hooks actions module."""

from dataclasses import dataclass

from .....actions import LoadAction, LoadData
from .....load.postgresql.transactions.operations import Operation, OperationType
from ....models.github import Repository, WebhookEvent


@dataclass
class RepoData(LoadData):
    """GitHub repository action data."""

    gh_repository: dict


class RepoCreateAction(LoadAction):
    """Create a GitHub repository."""

    name = "gh-repo-create"
    data_cls = RepoData

    def _generate_rows(self, **kwargs):
        """Generates rows for a gh repo create."""
        yield Operation(OperationType.INSERT, Repository, self.data.gh_repository)


class RepoUpdateAction(LoadAction):
    """Update a GitHub repository."""

    name = "gh-repo-update"
    data_cls = RepoData

    def _generate_rows(self, **kwargs):
        """Generates rows for a gh repo update."""
        # assumes the repo was created (INSERT) during repository sync or enable/create
        yield Operation(OperationType.UPDATE, Repository, self.data.gh_repository)


# NOTE: hook enabling happens in two phases, most likely due to multiple session commits
# the first assigned a hook id to the repository, the second updates the oauth server
# token and creates the actual webhook event.


@dataclass
class HookData(LoadData):
    """GitHub event hook action data."""

    webhook_event: dict


class HookEventCreateAction(LoadAction):
    """Create the webhook event in a GitHub repository."""

    name = "gh-hook-event-create"
    data_cls = HookData

    def _generate_rows(self, **kwargs):
        """Generates rows for the second phase of enabling a gh repo webhook."""
        yield Operation(OperationType.INSERT, WebhookEvent, self.data.webhook_event)


class HookEventUpdateAction(LoadAction):
    """Update the webhook event."""

    name = "gh-hook-event-update"
    data_cls = HookData

    def _generate_rows(self, **kwargs):
        """Generates rows for a webhook event update."""
        yield Operation(OperationType.UPDATE, WebhookEvent, self.data.webhook_event)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Base PostgreSQL generator."""

from abc import ABC, abstractmethod

from ...logging import Logger
from ...utils import dict_set


class PostgreSQLGenerator(ABC):
    """PostgreSQL generator interface.

    Prepares the data to be consumed by a Load class.
    """

    def __init__(self, pks=None, post_load_hooks=None):
        """Constructor."""
        self.post_load_hooks = post_load_hooks or []
        self.pks = pks or []

    def _generate_rows(self, **kwargs):
        """Yield generated rows."""
        # raises an error but does not force an implementation
        # e.g. when `prepare` is overwritten, _generate_rows is not required
        raise NotImplementedError

    def cleanup(self, **kwargs):
        """Cleanup."""
        pass

    def _generate_pks(self, data, create=False):
        keys = data.keys()
        for path, pk_func in self.pks:
            try:
                root = path.split(".")[0]
                # avoids creating e.g. "record" in a draft and generating a recid + uuid
                if create or root in keys:
                    dict_set(data, path, pk_func(data))
            except KeyError:
                logger = Logger.get_logger()
                logger.exception(f"Path {path} not found on record", exc_info=1)

    def _resolve_references(self, data, **kwargs):
        """Resolve references e.g communities slug names."""
        pass

    @abstractmethod
    def prepare(self, **kwargs):
        """Prepare data for loading."""
        pass

    def post_prepare(self, **kwargs):
        """Create rows after iterating over the entries."""
        pass

    def post_load(self, **kwargs):
        """Create rows after iterating over the entries."""
        for hook in self.post_load_hooks:
            hook(**kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Transaction load errors."""


class TableNotFound(Exception):
    """Table not found for migration."""

    def __init__(self, table_name):
        """Constructor."""
        super().__init__(
            f"Cannot process transaction row. Table not configured: {self.table_name}"
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""GitHub hooks actions module."""

from dataclasses import dataclass
from typing import Optional

import sqlalchemy as sa

from .....actions import LoadAction, LoadData
from .....load.ids import generate_recid, generate_uuid
from .....load.postgresql.transactions.operations import Operation, OperationType
from ....models.files import FilesBucket, FilesInstance, FilesObjectVersion
from ....models.github import Release, Repository
from ....models.pids import PersistentIdentifier
from ....models.records import (
    RDMParentMetadata,
    RDMRecordFile,
    RDMRecordMetadata,
    RDMVersionState,
)


@dataclass
class ReleaseData(LoadData):
    """GitHub release action data."""

    gh_release: dict
    gh_repository: Optional[dict] = None


class ReleaseReceiveAction(LoadAction):
    """Receive/Create a GitHub release action."""

    name = "gh-release-receive"
    data_cls = ReleaseData

    def _generate_rows(self, **kwargs):
        """Generates rows for a gh repo update."""
        assert self.data.gh_repository
        yield Operation(OperationType.UPDATE, Repository, self.data.gh_repository)
        yield Operation(OperationType.INSERT, Release, self.data.gh_release)


class ReleaseUpdateAction(LoadAction):
    """Update a GitHub release action."""

    name = "gh-release-update"
    data_cls = ReleaseData

    def _generate_rows(self, **kwargs):
        """Generates rows for a gh repo update."""
        yield Operation(OperationType.UPDATE, Release, self.data.gh_release)


@dataclass
class ReleaseProcessData(LoadData):
    """GitHub release action data."""

    record_pid: dict
    record_doi: dict
    record_oai: dict
    file_bucket: dict
    file_object: dict
    file_instance: dict
    file_record: dict
    parent: dict
    record: dict
    gh_release: dict

    parent_pid: Optional[dict] = None
    parent_doi: Optional[dict] = None


class ReleaseProcessAction(LoadAction):
    """Update a GitHub release action."""

    name = "gh-release-process"
    data_cls = ReleaseProcessData
    pks = [
        ("draft", "id", generate_uuid),
        ("parent", "id", generate_uuid),
        # TODO: decide if transform makes the json.pid or not, in drafts we do like that
        # not both parent.json.pid and draft.json.pid come filled in from the
        # transform action to match parent_pid and draft_pid data.
        # there is a TODO (to do it in the load - here -) in rdm-migrator/actions/drafts
        ("record", "json.pid", generate_recid),
        ("parent", "json.pid", generate_recid),
    ]

    def _generate_rows(self, session, **kwargs):
        """Generates rows for a gh repo update."""
        # pids

        # match gen pk
        self.data.record_pid["id"] = self.data.record["json"]["pid"]["pk"]
        yield Operation(
            OperationType.INSERT, PersistentIdentifier, self.data.record_pid
        )
        yield Operation(
            OperationType.INSERT, PersistentIdentifier, self.data.record_doi
        )
        yield Operation(
            OperationType.INSERT, PersistentIdentifier, self.data.record_oai
        )
        if self.data.parent_pid:  # first version
            self.data.parent_pid["id"] = self.data.parent["json"]["pid"]["pk"]
            yield Operation(
                OperationType.INSERT, PersistentIdentifier, self.data.parent_pid
            )
        if self.data.parent_doi:  # first version
            yield Operation(
                OperationType.INSERT, PersistentIdentifier, self.data.parent_doi
            )

        # bucket
        # no need to manage BUCKETS nor FILE_RECORD state
        # since after publishing the respective entries must be deleted from it
        # see drafts actions for clarification
        yield Operation(OperationType.INSERT, FilesBucket, self.data.file_bucket)
        # fi has a ref to fo so has to be inserted before
        yield Operation(OperationType.INSERT, FilesInstance, self.data.file_instance)
        yield Operation(OperationType.INSERT, FilesObjectVersion, self.data.file_object)
        yield Operation(OperationType.INSERT, RDMRecordFile, self.data.file_record)

        # records
        existing_parent_id = session.scalar(
            sa.select(PersistentIdentifier.object_uuid).where(
                PersistentIdentifier.pid_value == self.data.parent["json"]["id"],
                PersistentIdentifier.pid_type == "recid",
            )
        )
        parent_id = existing_parent_id or self.data.parent["id"]

        yield Operation(
            OperationType.DELETE,
            RDMRecordMetadata,
            dict(
                id=self.data.record["id"],
                json=self.data.record["json"],
                created=self.data.record["created"],
                updated=self.data.record["updated"],
                version_id=self.data.record["version_id"],
                index=self.data.record["index"],
                bucket_id=self.data.record["bucket_id"],
                parent_id=parent_id,
                expires_at=self.data.record["expires_at"],
                fork_version_id=self.data.record["fork_version_id"],
            ),
        )

        # parent and state
        if not existing_parent_id:
            yield Operation(
                OperationType.INSERT,
                RDMParentMetadata,
                dict(
                    id=parent_id,
                    json=self.data.parent["json"],
                    created=self.data.parent["created"],
                    updated=self.data.parent["updated"],
                    version_id=self.data.parent["version_id"],
                ),
            )

        # versioning
        versioning_op = (
            OperationType.UPDATE if existing_parent_id else OperationType.INSERT
        )
        yield Operation(
            versioning_op,
            RDMVersionState,
            dict(
                parent_id=self.data.parent["id"],
                latest_id=self.data.record["id"],
                latest_index=self.data.record["index"],
                next_draft_id=None,
            ),
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Requests transform tests."""

import dictdiffer

from invenio_rdm_migrator.streams.communities import CommunityEntry, CommunityTransform


class MockCommunityTransform(CommunityTransform):
    """Mock class for community data transformation."""

    def _community(self, entry):
        """Transform the community."""
        return {"key": "value"}

    def _oai_set(self, entry):
        """Transform the featured community."""
        return {"key": "value"}

    def _community_members(self, entry):
        """Tramsform the community members."""
        return {"key": "value"}

    def _featured_community(self, entry):
        """Transform the featured community."""
        return {"key": "value"}

    def _community_files(self, entry):
        return {
            "file": {"key": "value"},
            "bucket": {"key": "value"},
            "file_object": {"key": "value"},
        }


def test_community_transform():
    """Test the full community transformation."""
    result = MockCommunityTransform()._transform({})
    expected = {
        "community": {"key": "value"},
        "oai_set": {"key": "value"},
        "community_members": {"key": "value"},
        "featured_community": {"key": "value"},
        "community_files": {
            "file": {"key": "value"},
            "bucket": {"key": "value"},
            "file_object": {"key": "value"},
        },
    }
    assert not list(dictdiffer.diff(result, expected))


class MockCommunityEntry(CommunityEntry):
    """Transform a single mock community entry."""

    def _created(self, entry):
        """Transform community 'created' field."""
        return "2023-01-01 12:00:00.00000"

    def _updated(self, entry):
        """Transform community 'updated' field."""
        return "2023-01-31 12:00:00.00000"

    def _version_id(self, entry):
        """Transform community 'version_id' field."""
        return 1

    def _slug(self, entry):
        """Transform community 'slug' field."""
        return "migrator"

    def _files(self, entry):
        """Transform community 'files' field."""
        return None

    def _access(self, entry):
        """Transform community 'access' field."""
        return None

    def _bucket_id(self, entry):
        """Transform community 'bucket_id' field."""
        return 1

    def _deletion_status(self, entry):
        """Transform community 'deletion_status' field."""
        return "P"

    def _metadata(self, entry):
        """Transform community 'metadata' field."""
        return {
            "title": "Migrator community",
            "description": "Migrator testing community",
            "page": "",
            "curation_policy": "",
        }


def test_community_entry():
    """Test the community entry transformation."""
    result = MockCommunityEntry().transform({})
    expected = {
        "created": "2023-01-01 12:00:00.00000",
        "updated": "2023-01-31 12:00:00.00000",
        "version_id": 1,
        "slug": "migrator",
        "json": {
            "$schema": "local://communities/communities-v1.0.0.json",
            "files": None,
            "access": None,
            "metadata": {
                "title": "Migrator community",
                "description": "Migrator testing community",
                "page": "",
                "curation_policy": "",
            },
        },
        "bucket_id": 1,
        "deletion_status": "P",
    }

    assert not list(dictdiffer.diff(result, expected))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Record table generator utilities."""

import psycopg

from ...models.files import FilesObjectVersion


class InsertRecordFiles:
    """Table generator post-hook for inserting record files from object versions."""

    def __init__(self, record_model, file_model, bucket_fk="bucket_id"):
        """Constructor."""
        self.record_model = record_model
        self.file_model = file_model
        self.bucket_fk = bucket_fk

    def __call__(self, db_uri=None):
        """Inserts record files from buckets and object versions."""
        assert db_uri  # should have come from kwargs

        with psycopg.connect(db_uri) as conn:
            # the query needs to be split in 3 parts because the empty jsonb dict
            # would cause problems with the string formatting
            insert = f"""
                INSERT INTO {self.file_model.__tablename__} (
                    id, json, created, updated, version_id, key, record_id, object_version_id
                )
            """
            select = "SELECT gen_random_uuid(), '{}'::jsonb, record.created, record.updated, 1, fo.key, record.id, fo.version_id"
            from_and_join = f"""
                FROM {self.record_model.__tablename__} AS record
                INNER JOIN {FilesObjectVersion.__tablename__} AS fo
                ON record.{self.bucket_fk} = fo.bucket_id AND fo.is_head = 'true'
                WHERE record.{self.bucket_fk} IS NOT NULL
            """
            # no return check, will raise if the sql statement fails
            # id and json do not have defaults in DB even though if the programmatic
            # models have them, so they need to be calculated
            conn.execute(insert + select + from_and_join)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Streams tests."""


from invenio_rdm_migrator.streams.streams import StreamDefinition


def test_stream_definition_attributes():
    # this tests is not super useful, just checks the attributes
    # imo, if it wasn't because it holds classes it could be a dataclass
    stream_def = StreamDefinition("test_stream", "extract", "transform", "load")
    assert stream_def.name == "test_stream"
    assert stream_def.extract_cls == "extract"
    assert stream_def.transform_cls == "transform"
    assert stream_def.load_cls == "load"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration users stream."""

from .load import UserCopyLoad
from .transform import UserEntry, UserTransform

__all__ = (
    "UserCopyLoad",
    "UserEntry",
    "UserTransform",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration transform interfaces."""

from abc import ABC, abstractmethod
from typing import Optional, Union

import pypeln

from ..logging import Logger


class Transform(ABC):
    """Base class for data transformation."""

    def __init__(self, workers=None, throw=False):
        """Initialize base transform."""
        self._workers = workers
        self._throw = throw

    @abstractmethod
    def _transform(self, entry):  # pragma: no cover
        """Transform entry.

        :returns: a transformed entry, not an iterator.
        """
        pass

    def _multiprocess_transform(self, entries):
        def _transform(entry):
            try:
                return self._transform(entry)
            except Exception:
                logger = Logger.get_logger()
                logger.exception(entry, exc_info=True)
                if self._throw:
                    raise

        for result in pypeln.process.map(
            _transform,
            entries,
            workers=self._workers,
            maxsize=self._workers * 100,
        ):
            if result:
                yield result

    def run(self, entries):
        """Transform and yield one element at a time."""
        if self._workers is None:
            for entry in entries:
                try:
                    yield self._transform(entry)
                except Exception:
                    logger = Logger.get_logger()
                    logger.exception(entry, exc_info=True)
                    if self._throw:
                        raise
                    continue
        else:
            yield from self._multiprocess_transform(entries)


class Entry(ABC):
    """Base entry class."""

    def __init__(self, partial=False):
        """Constructor.

        :param partial: a boolean enabling partial transformations, i.e. missing keys.
        """
        self.partial = partial

    def _load_partial(
        self,
        entry: dict,
        obj: dict,
        keys: list[Union[str, tuple[str, str]]],
        prefix: Optional[str] = None,
    ):
        for key in keys:
            if isinstance(key, tuple):
                key, func_key = key
            else:
                func_key = key
            func = getattr(self, "_" + func_key)
            try:
                val = func(entry)
                if prefix:
                    obj.setdefault(prefix, {})
                    obj[prefix][key] = val
                else:
                    obj[key] = val
            # this might mask nested missing keys, it is still a partial transformation
            # full one (with more validation) should be checked on a record
            except KeyError as ex:
                if not self.partial:
                    raise KeyError(ex)
                pass

    @abstractmethod
    def transform(self, entry):  # pragma: no cover
        """Transform entry."""
        pass


def drop_nones(data):
    """Recursively drop Nones in dict d and return a new dictionary."""
    dd = {}
    for k, v in data.items():
        if isinstance(v, dict) and v:  # second clause removes empty dicts
            dd[k] = drop_nones(v)
        elif v is not None:
            dd[k] = v
    return dd



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Existing data table generator module."""

from .table import TableGenerator


class ExistingDataTableGenerator(TableGenerator):
    """Table generator to import data directly from existing data.

    This is useful when the data_dir already contain the files (e.g. csv) with
    the contents to be imported, for example from a previous migration run. Using
    this table generator the Extract and Transform steps can be skipped.
    """

    def __init__(self, tables, pks=None, post_load_hooks=None):
        """Constructor."""
        # forces pks to None since they won't be used
        super().__init__(
            tables, pks=None, post_load_hooks=post_load_hooks, existing_data=True
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Dataclasses models to generate table rows for OAI-PMH."""

from dataclasses import InitVar
from datetime import datetime

from sqlalchemy.orm import Mapped, mapped_column

from ...load.postgresql.models import Model


class OAISet(Model):
    """OAI set dataclass model."""

    __tablename__: InitVar[str] = "oaiserver_set"

    id: Mapped[int] = mapped_column(primary_key=True)
    spec: Mapped[str] = mapped_column(nullable=False)
    name: Mapped[str] = mapped_column(nullable=True)
    description: Mapped[str] = mapped_column(nullable=True)
    search_pattern: Mapped[str] = mapped_column(nullable=True)
    system_created: Mapped[bool] = mapped_column(nullable=False)

    created: Mapped[datetime]
    updated: Mapped[datetime]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Records state module."""

from ...state import StateValidator


class ParentModelValidator(StateValidator):
    """Parent state entry validator."""

    @classmethod
    def validate(cls, data):
        """Data validation."""
        latest_id = data.get("latest_id")
        assert latest_id or data.get("next_draft_id")
        if latest_id:
            assert data.get("latest_index")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Base load tests."""

import pytest

from invenio_rdm_migrator.load import Load


def test_cannot_instantiate_base_class():
    with pytest.raises(TypeError):
        Load()

    class Test(Load):
        """Test Load."""

        pass

    with pytest.raises(TypeError):
        Test()


class BaseTestLoad(Load):
    """Test load implementation."""

    def _prepare(self, entry):
        pass

    def _load(self, entry):
        pass

    def _cleanup(self):
        pass


def test_entries_loaded_successfully():
    test_load_impl = BaseTestLoad()
    test_load_impl.run([1, 2, 3])


def test_skip_loading_invalid_entries():
    class TestLoadImpl(BaseTestLoad):
        def __init__(self):
            self.seen = []

        def _prepare(self, entry):
            self.seen.append(entry)

        def _load(self, entry):
            assert entry in self.seen

        def _validate(self, entry):
            return entry != 2

    test_load_impl = TestLoadImpl()
    test_load_impl.run([1, 2, 3])
    assert test_load_impl.seen == [1, 3]


def test_cleanup_called():
    class TestLoadImpl(BaseTestLoad):
        def __init__(self):
            self.cleanup = None

        def _cleanup(self):
            self.cleanup = True

    test_load_impl = TestLoadImpl()
    test_load_impl.run([])
    assert not test_load_impl.cleanup
    test_load_impl.run([], cleanup=True)
    assert test_load_impl.cleanup



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Base extract tests."""

import pytest

from invenio_rdm_migrator.extract import Extract


def test_cannot_instantiate_base_class():
    with pytest.raises(TypeError):
        Extract()

    class Test(Extract):
        """Test Extract."""

        pass

    with pytest.raises(TypeError):
        Test()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""PostgreSQL COPY load."""


import contextlib
from dataclasses import fields
from pathlib import Path

import psycopg

from ....logging import Logger
from ....utils import ts
from ...base import Load
from ..sequences import AlterSequencesMixin


class PostgreSQLCopyLoad(Load, AlterSequencesMixin):
    """PostgreSQL COPY load."""

    def __init__(
        self,
        db_uri,
        table_generators,
        tmp_dir=None,
        data_dir=None,
        existing_data=False,
        **kwargs,
    ):
        """Constructor.

        :param data_dir: if existing data is true this is the directory from which to
        load the existing csv file, if it is false is the directory where to dump the
        newly created csv files.
        """
        self.db_uri = db_uri
        self.table_generators = table_generators
        # when loading existing data the tmp folder would be the root
        # it is assumed that the csv files of a previous run have been placed there
        self.existing_data = existing_data
        self.data_dir = None
        self.tmp_dir = None
        if existing_data:
            assert data_dir

        assert data_dir or tmp_dir

        if data_dir:
            self.data_dir = Path(data_dir)

        if tmp_dir:
            self.tmp_dir = Path(tmp_dir) / f"tables-{ts(fmt='%Y-%m-%dT%H%M%S')}"

    def _cleanup(self, db=False):
        """Cleanup csv files and DB after load."""
        for table in self.table_generators:
            table.cleanup(db=db)

    def _save_to_csv(self, entries):
        """Save the entries to a csv file."""
        self.tmp_dir.mkdir(parents=True, exist_ok=True)
        # use this context manager to close all opened files at once
        with contextlib.ExitStack() as stack:
            output_files = {}
            for entry in entries:
                for tg in self.table_generators:
                    tg.prepare(self.tmp_dir, entry, stack, output_files)

            for tg in self.table_generators:
                tg.post_prepare(
                    tmp_dir=self.tmp_dir, stack=stack, output_files=output_files
                )

    def _prepare(self, entries):
        """Dump entries in csv files for COPY command."""
        # global overwrite for existing data, e.g. when running a previously run stream
        if not self.existing_data:
            self._save_to_csv(entries)

        prepared_tables = []
        # Needs to preserve order. This logic is very inefficient, maybe and ordered-set
        # or change the data structure of the tables to keep order information and then
        # process it
        for tg in self.table_generators:
            for table in tg.tables:
                existing_data = tg.existing_data or self.existing_data
                if table not in prepared_tables:
                    prepared_tables.append((existing_data, table))

        return iter(prepared_tables)  # yield at the end vs yield per table

    def _load(self, table_entries):
        """Bulk load CSV table files.

        Loads the tables in the order given by the generator.
        """
        logger = Logger.get_logger()

        with psycopg.connect(self.db_uri) as conn:
            for existing_data, table in table_entries:
                name = table.__tablename__
                cols = ", ".join([f.name for f in fields(table)])
                # local overwrite for existing data
                # e.g. when a table does not need transformation and is already in csv
                fpath = self.data_dir if existing_data else self.tmp_dir
                fpath = fpath / f"{name}.csv"

                if fpath.exists():
                    # total file size for progress logging
                    file_size = fpath.stat().st_size

                    logger.info(f"COPY FROM {fpath}.")
                    with contextlib.ExitStack() as stack:
                        cur = stack.enter_context(conn.cursor())
                        copy = stack.enter_context(
                            cur.copy(f"COPY {name} ({cols}) FROM STDIN (FORMAT csv)")
                        )
                        fp = stack.enter_context(open(fpath, "r"))

                        block_size = 8192

                        def _data_blocks(block_size=8192):
                            data = fp.read(block_size)
                            while data:
                                yield data
                                data = fp.read(block_size)

                        for idx, block in enumerate(_data_blocks(block_size)):
                            if idx % 100:
                                cur_bytes = idx * block_size
                                percentage = (cur_bytes / file_size) * 100
                                progress = (
                                    f"{cur_bytes}/{file_size} ({percentage:.2f}%)"
                                )
                                logger.info(f"{name}: {progress}")
                            copy.write(block)
                else:
                    logger.warning(f"{name}: no data to load.")
                conn.commit()

    def _post_load(self):
        """Post load processing."""
        self.alter_sequences()

    def run(self, entries, cleanup=False):
        """Load entries."""
        table_entries = self._prepare(entries)
        self._load(table_entries)
        self._post_load()

        if cleanup:
            self._cleanup()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""GitHub actions module."""

from .hooks import (
    HookEventCreateAction,
    HookEventUpdateAction,
    RepoCreateAction,
    RepoUpdateAction,
)
from .releases import ReleaseProcessAction, ReleaseReceiveAction, ReleaseUpdateAction

__all__ = (
    "HookEventCreateAction",
    "HookEventUpdateAction",
    "RepoCreateAction",
    "RepoUpdateAction",
    "ReleaseReceiveAction",
    "ReleaseUpdateAction",
    "ReleaseProcessAction",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Files actions module."""

from copy import deepcopy
from dataclasses import dataclass
from typing import Optional

import sqlalchemy as sa

from ....actions import LoadAction, LoadData
from ....load.ids import generate_uuid
from ....load.postgresql.transactions.operations import Operation, OperationType
from ....state import STATE
from ...models.files import FilesBucket, FilesInstance, FilesObjectVersion
from ...models.records import RDMDraftFile, RDMDraftMetadata


@dataclass
class FileUploadData(LoadData):
    """File upload action data."""

    bucket: dict
    object_version: dict
    file_instance: dict
    file_record: dict
    replaced_object_version: Optional[dict] = None


class FileUploadAction(LoadAction):
    """File upload action."""

    name = "file-upload"
    data_cls = FileUploadData
    data: FileUploadData
    pks = [("file_record", "id", generate_uuid)]

    def _generate_rows(self, session, **kwargs):
        """Generates rows for a new draft."""
        replaced_ov = self.data.replaced_object_version
        new_ov = self.data.object_version
        yield Operation(OperationType.UPDATE, FilesBucket, self.data.bucket)
        if not replaced_ov:
            replaced_ov_model = session.execute(
                sa.select(FilesObjectVersion.version_id).where(
                    FilesObjectVersion.bucket_id == self.data.bucket["id"],
                    FilesObjectVersion.key == new_ov["key"],
                    FilesObjectVersion.is_head.is_(True),
                )
            ).one_or_none()
            if replaced_ov_model:
                replaced_ov = {"version_id": replaced_ov_model.version_id}

        if replaced_ov:
            yield Operation(
                OperationType.UPDATE,
                FilesObjectVersion,
                {"version_id": replaced_ov["version_id"], "is_head": False},
            )

        yield Operation(OperationType.INSERT, FilesInstance, self.data.file_instance)
        yield Operation(OperationType.INSERT, FilesObjectVersion, new_ov)
        self.data.file_record["object_version_id"] = new_ov["version_id"]
        if replaced_ov:
            file_record_id = session.execute(
                sa.select(RDMDraftFile.id).where(
                    RDMDraftFile.object_version_id == replaced_ov["version_id"],
                )
            ).scalar_one()
            self.data.file_record["id"] = file_record_id
            yield Operation(OperationType.UPDATE, RDMDraftFile, self.data.file_record)
        else:
            # Find draft ID
            draft_id = session.execute(
                sa.select(RDMDraftMetadata.id).where(
                    RDMDraftMetadata.bucket_id == self.data.bucket["id"]
                )
            ).scalar_one()
            self.data.file_record["record_id"] = draft_id
            yield Operation(OperationType.INSERT, RDMDraftFile, self.data.file_record)


@dataclass
class FileDeleteData(LoadData):
    """File delete action data."""

    bucket: dict
    deleted_object_version: dict
    delete_marker_object_version: Optional[dict] = None


class FileDeleteAction(LoadAction):
    """File delete action."""

    name = "file-delete"
    data_cls = FileDeleteData
    data: FileDeleteData

    def _generate_rows(self, *, session, **kwargs):
        """Generates rows for deleting a draft file."""
        bucket = self.data.bucket
        deleted_ov = self.data.deleted_object_version
        delete_marker = self.data.delete_marker_object_version

        yield Operation(OperationType.UPDATE, FilesBucket, bucket)

        # We always soft delete, which means we update the old OV's is_head = False...
        yield Operation(OperationType.UPDATE, FilesObjectVersion, deleted_ov)
        # ...and insert a delete marker on top
        if not delete_marker:
            delete_marker = {
                "version_id": generate_uuid(),
                "bucket_id": bucket["id"],
                "key": deleted_ov["key"],
                "created": bucket["updated"],
                "updated": bucket["updated"],
                "is_head": True,
            }
            yield Operation(OperationType.INSERT, FilesObjectVersion, delete_marker)

        # Finally we delete the draft file
        file_record = session.scalars(
            sa.select(RDMDraftFile).where(
                RDMDraftFile.object_version_id == deleted_ov["id"],
                RDMDraftFile.key == deleted_ov["key"],
            )
        ).one_or_none()
        if file_record:
            yield Operation(OperationType.DELETE, RDMDraftFile, {"id": file_record.id})


@dataclass
class MediaFileUploadData(LoadData):
    """Media file upload action data."""

    record: dict
    bucket: dict
    object_version: dict
    file_instance: dict
    file_record: dict
    replaced_object_version: Optional[dict]


class MediaFileUploadAction(LoadAction):
    """Media file upload action."""

    name = "media-file-upload"
    data_cls = MediaFileUploadData
    data: MediaFileUploadData
    pks = [("file_record", "id", generate_uuid)]

    def _generate_rows(self, **kwargs):
        """Generates rows for a new draft."""
        # if we were to use the state for consistency checks
        # the bucket should already exist
        yield Operation(OperationType.UPDATE, FilesBucket, self.data.bucket)
        yield Operation(OperationType.INSERT, FilesInstance, self.data.file_instance)
        yield Operation(
            OperationType.INSERT, FilesObjectVersion, self.data.object_version
        )

        cached_bucket = STATE.BUCKETS.get(self.data.bucket["id"])
        self.data.file_record["record_id"] = cached_bucket["draft_id"]

        _cache_fr_data = deepcopy(self.data.file_record)
        STATE.FILE_RECORDS.add(_cache_fr_data.pop("id"), _cache_fr_data)

        yield Operation(OperationType.INSERT, RDMDraftFile, self.data.file_record)


@dataclass
class MediaFileDeleteData(LoadData):
    """Media file delete action data."""

    bucket: dict
    object_version: dict
    file_instance: dict
    file_record: dict


class MediaFileDeleteAction(LoadAction):
    """Media file delete action."""

    name = "media-file-delete"
    data_cls = MediaFileDeleteData
    data: MediaFileDeleteData

    def _generate_rows(self, **kwargs):
        """Generates rows for a new draft."""
        # if we were to use the state for consistency checks
        # the bucket should already exist
        yield Operation(OperationType.UPDATE, FilesBucket, self.data.bucket)
        yield Operation(OperationType.INSERT, FilesInstance, self.data.file_instance)
        yield Operation(
            OperationType.INSERT, FilesObjectVersion, self.data.object_version
        )

        cached_bucket = STATE.BUCKETS.get(self.data.bucket["id"])
        self.data.file_record["record_id"] = cached_bucket["draft_id"]

        _cache_fr_data = deepcopy(self.data.file_record)
        STATE.FILE_RECORDS.add(_cache_fr_data.pop("id"), _cache_fr_data)

        yield Operation(OperationType.INSERT, RDMDraftFile, self.data.file_record)
